//Implementation of an active MemPool
//Each time Memory is freed the class checks whether there are preceding or following memory blocks
//that can be connected to one bigger block. Free blocks are then linked internally.
//It is crucial for this class that used memory blocks must not start or end with data 0xFFFE or 0xFFFF.
//For an allocated block [||] a user can write [23|0xFFFF...|42] but not [23|42...|0xFFFF] or [0xFFFF|...23|42]
#ifndef MEMPOOL_HPP_
#define MEMPOOL_HPP_
#define mempoola_hpp_version	  "$Revision: 1.4 $"

#include<vector>
#include"Assertion.hpp"
#include "DbgAlloc.hpp"
#include "Logging.hpp"
#include "BitVector.hpp"
#include "Ptr32.hpp"
#include <stdlib.h>

using std::vector;

//#define REPRO(x,w) {cout << x <<" " << w << " id " << ID << endl;}
#define REPRO(x,w) asrt(true);


/**Organisation:
 *
 * There is a strong restriction for the use of this MemPool class:
 * It has to be guaranteed from outside that memory allocated by this MemPool
 * have the first 4 bytes and the last 4 bytes smaller than 2^32 - 2 (max32-1).
 * The minimum memory that can be allocated as to be at least 3 units (1 unit is always 4 bytes).
 * Freed blocks are marked internally as follows:
 * Freed blocks of 1 units: 	[FREED_SMALL]
 * Freed blocks of 2 units: 	[FREED_SMALL;FREED_SMALL]
 * Freed blocks of 3 units: 	[FREED_BLOCK;index in vector freed_trns;FREED_BLOCK]
 * Freed blocks of 4 units: 	[FREED_BLOCK;link to previous in freed list;link to next in freed list;FREED_BLOCK]
 * Freed blocks of 5 units: 	[FREED_BLOCK;link to previous in freed list;link to next in freed list;size of block;FREED_BLOCK]
 * Freed blocks of >= 6 units:	[FREED_BLOCK;link to previous in freed list;link to next in freed list;size of block; ... unused ... ;size of block;FREED_BLOCK]
 *
 * There are different linked lists of freed blocks for different sizes (for size > 3) referenced from vector freed_blks.
 * All blocks with size in range [2^i,2^i+1) are linked in one list.
 * Freed blocks of size 3 are referenced from the vector freed_trns
 * Freed blocks of size < 3 are not linked.
 * Freed blocks are always mereged to be as large as possible. Two freed blocks can never be neighboured. *
 */
class MemPoolA{
	static const uint32_t FREED_SMALL	= max32; //for free blocks of size 1,2
	static const uint32_t FREED_BLOCK 	= max32-1;//for free blocks of size >= 3
	static const uint32_t EMPTY_LIST	= max32;

	//actual memory where user data is stored

	uint32_t* 	mem;
	uint32_t	mem_end;
	uint32_t	mem_sz;
	uint32_t 	frgmnts_sum;
	const char  ID;

	//heads of double linked lists to freed blocks of size >= 4 (ints)
	vector<uint32_t>	freed_blks; //an entry at index i is a head of a list of blocks with size >= 2^(i+2) and < 2^(i+3)

	//indices of freed blocks of size 3 (ints)
	vector<uint32_t>	freed_trns;


	//vector<uint32_t>	dbg_used;

	MemPoolA(const MemPoolA& m);
	MemPoolA& operator=(const MemPoolA& m);


public:
	inline void hardClear();
	MemPoolA(const char id):mem(NULL),mem_end(0),mem_sz(0),frgmnts_sum(0),ID(id){}
	~MemPoolA(){hardClear();}
	uint64_t getMemTotal() const{
		return (mem_sz + freed_blks.size() + freed_trns.size()) << 2;}

	inline uint32_t getFragmentsSum() const{return frgmnts_sum;}
	inline bool	inRange(const Ptr32& p) const{return p.idx() < mem_sz;}

	inline void increaseTo(const uint64_t req, const uint64_t min_sz);
	void reserveFor(const uint32_t sz){LOG3("Reserve mempool for "<<sz << " id " << ID);
		if(sz<=mem_end) return; increaseTo(sz,sz);}

	template<class C>
	C* getRealPtr(const Ptr32& p)	{asrt(!p.isNull()); asrt(p.id<mem_sz);
									 return (C*)&mem[p.id];}

	template<class C>
	C& getRef(const Ptr32& p)		{return (C&)*(&mem[p.id]);}

	template<class C>
	const C& getRef(const Ptr32& p) const {return (C&)*(&mem[p.id]);}

	void copy(const Ptr32& from, const Ptr32& to){asrt(!from.isNull()); asrt(!to.isNull()); asrt(from.id<mem_sz); asrt(to.id<mem_sz);
		mem[to.id]=mem[from.id];}

	inline void dealloc(const Ptr32& beg, const Ptr32& end);
	inline Ptr32 alloc(const uint32_t sz);
	inline uint32_t tryAllocAt(const Ptr32& p, const uint32_t& min_sz, const uint32_t& max_sz);
	void dealloc(const Ptr32& beg, const uint32_t& sz){asrt(sz!=0);dealloc(beg,beg+sz);}

	bool validEndData(const Ptr32& at)const {return at.id < mem_sz &&
			(mem[at.id] < FREED_BLOCK);}


private:

	inline void	freeUnextendable(const uint32_t& beg, const uint32_t& end);
	inline void addToFreedList(const uint32_t blk_hd, const uint32_t blk_end);
	inline void removeFromFreedList(const uint32_t blk_hd, const uint32_t blk_end);
	inline void addTernToFreedList(const uint32_t blk_hd);
	inline void removeTernFromFreedList(const uint32_t blk_hd);


	//helpers for freed blocks with size > 3
	//all functions are used to mangae linkage of deleted blocks
	//functions delFr* access freed blocks from front, functions delBk* access freed blocks from back
	uint32_t& delFrBlkBegFld(const uint32_t& hd) 		{asrt(hd+2<mem_sz); return mem[hd];}
	uint32_t& delFrBlkSzFld(const uint32_t& hd) 		{return mem[hd+3];}
	uint32_t  delFrBlkSize(const uint32_t& hd) const	{asrt(hd+3<mem_sz); asrt(!delFrIsTern(hd));
														 uint32_t sz = mem[hd+3]; return sz==FREED_BLOCK ? 4 : sz;}
	uint32_t& delFrBlkPrev(const uint32_t& hd) 			{return mem[hd+1];}
	uint32_t& delFrBlkNext(const uint32_t& hd) 			{return mem[hd+2];}
	uint32_t& delBkBlkEndFld(const uint32_t& end)		{asrt(end>2); return mem[end-1];}

	uint32_t& delBkBlkSzFld(const uint32_t& end)		{asrt(end>3); return mem[end-2];}
	uint32_t  delBkBlkSize(const uint32_t& end)const	{asrt(end>3); return mem[end-4]==FREED_BLOCK ? 4 : mem[end-2];}

	//helpers for blocks with size 3
	bool	  delFrIsTern(const uint32_t& hd)	const	{asrt(hd+2<mem_sz); return mem[hd+2]==FREED_BLOCK;}
	uint32_t& delFrTernEndFld(const uint32_t& hd)		{asrt(hd+2<mem_sz); return mem[hd+2];}
	uint32_t& delFrTernIdx(const uint32_t& hd)			{asrt(hd+2<mem_sz); return mem[hd+1];}
	bool	  delBkIsTern(const uint32_t& end)	const	{asrt(end>2); return mem[end-3]==FREED_BLOCK;}

	void	  delClearControlFlags(const uint32_t& hd, const uint32_t& end){
				asrt(end-hd>=3);mem[hd]=mem[hd+1]=mem[hd+2]=mem[end-1]=mem[end-2]=0; }


	uint32_t sizeToFloorIndex(uint32_t v){ asrt(v >= 4);
		uint32_t l = 0; v>>=3;
		while(v){v >>= 1;++l;}
		return l;
	}

	uint32_t sizeToCeilIndex(uint32_t v){ asrt(v >= 3);
		uint32_t l = 0; v = (v-1) >> 2;
		while(v){v >>= 1;++l;}
		return l;
	}

};


void 	MemPoolA::removeFromFreedList(const uint32_t blk_hd, const uint32_t blk_end){
	////LOG_ALWAYS("removeFromFreedList memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);

	asrt(blk_hd+4 <= mem_sz); asrt(blk_end-blk_hd>=4);asrt(delFrBlkBegFld(blk_hd)==FREED_BLOCK); asrt(delBkBlkEndFld(blk_end)==FREED_BLOCK);
	uint32_t size			= blk_end-blk_hd; asrt(frgmnts_sum>=size);
	frgmnts_sum				-= size;
	uint32_t next			= delFrBlkNext(blk_hd);
	uint32_t prev			= delFrBlkPrev(blk_hd);
	delClearControlFlags(blk_hd,blk_end);

	//notify previous if there is one
	if(prev==EMPTY_LIST){ //first in list
		uint32_t lidx		= sizeToFloorIndex(size);
								asrt(lidx<freed_blks.size()); asrt(freed_blks[lidx]==blk_hd);
		freed_blks[lidx]	= next;

	}else{						asrt(prev+3<mem_sz); asrt(delFrBlkBegFld(prev)==FREED_BLOCK);
		delFrBlkNext(prev)	= next;
	}

	if(next != EMPTY_LIST){asrt(delFrBlkBegFld(next)==FREED_BLOCK);
		delFrBlkPrev(next)	= prev;}

}


void 	MemPoolA::addToFreedList(const uint32_t blk_hd, const uint32_t blk_end){ asrt(blk_end>blk_hd);
	////LOG_ALWAYS("addToFreedList memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);

	uint32_t size 			= blk_end - blk_hd; asrt(size >=4); asrt(blk_end <= mem_sz);
	frgmnts_sum 			+= size;
	uint32_t lidx			= sizeToFloorIndex(size);

	if(lidx >= freed_blks.size()){
		freed_blks.resize(lidx+1,EMPTY_LIST);}

	uint32_t next				= freed_blks[lidx];
	mem[blk_end-4]				= 0;	//may be overwritten -> need to be cleared for delBkBlkSz(blk_end) calls
	mem[blk_end-3]				= 0;	//may be overwritten -> need to be cleared for delBkIsTern(blk_end) calls
	delFrBlkSzFld	(blk_hd)	= size; // may be overwritten if size < 6
	delBkBlkSzFld	(blk_end)	= size; // may be overwritten if size < 5
	delFrBlkBegFld	(blk_hd)	= FREED_BLOCK;
	delBkBlkEndFld	(blk_end)	= FREED_BLOCK;
	delFrBlkNext	(blk_hd)	= next;
	delFrBlkPrev	(blk_hd)	= EMPTY_LIST; //no prev

	if(next != EMPTY_LIST)	{asrt(delFrBlkBegFld(next)==FREED_BLOCK);
			delFrBlkPrev(next)	= blk_hd;}

	freed_blks[lidx]			= blk_hd;
}

void MemPoolA::addTernToFreedList(const uint32_t blk_hd){asrt(blk_hd+2<mem_sz);
	////LOG_ALWAYS("addTernToFreedList memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);

	delFrBlkBegFld(blk_hd) 		= FREED_BLOCK;
	delFrTernIdx(blk_hd)		= freed_trns.size(); //index into freed_trns
	delFrTernEndFld(blk_hd)		= FREED_BLOCK;
	freed_trns					.push_back(blk_hd);
	frgmnts_sum 				+= 3;
}

void MemPoolA::removeTernFromFreedList(const uint32_t blk_hd){asrt(blk_hd+2<mem_sz);
	////LOG_ALWAYS("removeTernFromFreedList memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);

	asrt(mem[blk_hd]==FREED_BLOCK); asrt(mem[blk_hd+2]==FREED_BLOCK); asrt(!freed_trns.empty());

	const uint32_t idx 		= delFrTernIdx(blk_hd); asrt(idx < freed_trns.size()); asrt(freed_trns[idx]==blk_hd);
	delFrBlkBegFld(blk_hd)	= 0;
	delFrTernEndFld(blk_hd)	= 0; asrt(frgmnts_sum>=3);
	frgmnts_sum 			-= 3;

	//move last in freed_trns to this position
	const uint32_t last		= freed_trns.back();
	delFrTernIdx(last)		= idx;
	freed_trns[idx]			= last;
	freed_trns				.pop_back();
}


void 	MemPoolA::dealloc(const Ptr32& beg, const Ptr32& end){
	////LOG_ALWAYS("dealloc memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);

	LOG3("Free mem from " << beg.idx() << " to " << end.idx() << " and max range is " << mem_sz<< " id " << ID);
	asrt(!beg.isNull()); asrt(!end.isNull()); asrt(end.id <=mem_sz); asrt(beg.id!=end.id); asrt(beg.id<end.id);
	REPRO('D', beg.idx() << " " << end.idx());

	uint32_t dbeg	= beg.id;
	uint32_t dend 	= end.id;
	//bool ok = markFreed(dbeg, dend-dbeg);asrt(ok);

	//connect with preceding block if freed
	if(dbeg>0){
		//asrt(mem[dbeg-1]!=UN_INIT);

		//check if small unlinked block (<3) is preceding
		//cerr << "check mem at " << dbeg-1 << endl;
		if(mem[dbeg-1]==FREED_SMALL){
			mem[dbeg-1]	= 0;		LOG3("Extend freed to "<< dbeg-1<< " id " << ID);

			dbeg -= 1 + (dbeg>1 && mem[dbeg-2]==FREED_SMALL);
			mem[dbeg]	= 0; //may be same as above


		//check if bigger block is preceding
		}else if(mem[dbeg-1]==FREED_BLOCK){ asrt(dbeg>=3);

			//block of size 3 is preceding
			if(delBkIsTern(dbeg)){
				dbeg			-= 3;
				removeTernFromFreedList(dbeg);
				LOG3("Extend freed prev to "<< dbeg << " " << dbeg+1 << " and "<< dbeg+2<< " id " << ID);

			//block of size > 3 is preceding
			}else{
				uint32_t pr_end	= dbeg; //current dbeg is end of preceding block
				dbeg			-= delBkBlkSize(pr_end); asrt(dbeg < pr_end);
				removeFromFreedList(dbeg,pr_end);
			}
		}
	}

	//shrink data vector if this block is at the end of mem
	if(dend>=mem_sz){
		asrt(dbeg<mem_sz);LOG3("Shrink mem at end"<< " id " << ID);
		mem_sz = dbeg;
		return;
	}


	//asrt(mem[dend]!=UN_INIT);
	//connect with next block if freed
	if(mem[dend]==FREED_SMALL){
		asrt(dend+1<mem_sz); LOG3("Extend freed post to "<< dend+1<< " id " << ID);
		mem[dend]	= 0;

		dend += 1 + (mem[dend+1]==FREED_SMALL);
		mem[dend-1]	= 0; //may be same as above


	}else if(mem[dend]==FREED_BLOCK){

		//freed block of size 3 is next
		if(delFrIsTern(dend)){
			removeTernFromFreedList(dend);
			LOG3("Extend freed post to "<< dend << " " << dend+1 << " and "<< dend+2<< " id " << ID);
			dend			+= 3;


		//freed block of size > 3 is next
		}else{
			uint32_t next_beg	= dend;
			dend				+= delFrBlkSize(next_beg);
			removeFromFreedList(next_beg,dend);
		}
	}

	freeUnextendable(dbeg,dend);
}

void  MemPoolA::freeUnextendable( const uint32_t& beg, const uint32_t& end){
	////LOG_ALWAYS("freeUnextendable memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);
	LOG3("Free unextendable from " << beg << " to " << end <<  " id " << ID);
	const uint32_t sz = end-beg; asrt(sz>0); asrt(end > beg);

	if(sz > 3){
		addToFreedList(beg,end);
		return;}

	if(sz == 3){
		addTernToFreedList(beg);
		return;}

	mem[beg]	= FREED_SMALL;
	mem[end-1]	= FREED_SMALL;
}

uint32_t MemPoolA::tryAllocAt(const Ptr32& p, const uint32_t& min_sz, const uint32_t& max_sz){
	////LOG_ALWAYS("tryAllocAt memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);
	asrt(min_sz>0); asrt(max_sz>=min_sz); asrt(max_sz>=3); LOG3("Try allocate mem at position " << p.idx() << " of size " << min_sz<< " id " << ID);


	//extending memory would require to reallocate mem
	if((uint64_t)p.id + (uint64_t)min_sz > (uint64_t)mem_sz ) return 0;

	if(p.id == mem_sz) {
		LOG3("Extend mempool at end by tryAlloc"<< " id " << ID);
		mem_sz += min_sz; asrt(mem_sz<= mem_end);
		delClearControlFlags(p.id,p.id+min_sz);
		return min_sz;
	}


	asrt(p.id<mem_sz); //asrt(mem[p.id]!=UN_INIT);
	if(mem[p.id] == FREED_SMALL){ asrt(p.id+1<mem_sz);
		uint32_t sz = 1 + (mem[p.id+1]==FREED_SMALL);
		if(sz < min_sz) return 0;
		mem[p.id]		= 0;
		mem[p.id+sz-1]	= 0;//may be same
		return sz;
	}

	if(mem[p.id] != FREED_BLOCK){ return 0;}

	if(delFrIsTern(p.id)){
		if(3<min_sz) return 0;

		removeTernFromFreedList(p.id);
		return 3;
	}

	uint32_t real_sz = delFrBlkSize(p.id);
	if(real_sz<min_sz) return 0;

	removeFromFreedList(p.id,p.id+real_sz);

	if(real_sz<=max_sz){return real_sz;	}

	freeUnextendable(p.id+max_sz,p.id+real_sz);

	delClearControlFlags(p.id,p.id+max_sz);
	return max_sz;
}


Ptr32 MemPoolA::alloc(const uint32_t sz){
	asrt(sz>=3); LOG3("Start allocating memory mem_sz is " << mem_sz << " mem_end is " << mem_end <<" id " << ID);
	uint32_t hd;

	if(sz==3 && !freed_trns.empty()){
		hd 					= freed_trns.back(); freed_trns.pop_back();
		delFrBlkBegFld(hd)	= 0;
		delFrTernEndFld(hd)		= 0;	LOG3("1. Alloc mem from " << hd << " to " << hd+3<< " mem_sz is " << mem_sz << " mem_end is " << mem_end <<" id " << ID);
		REPRO('A',sz << " at " << hd);//bool ok = ensureUnused(hd,sz); asrt(ok);
		return Ptr32(hd);
	}

	const uint32_t ceil		= sizeToCeilIndex(sz);
	for(uint32_t lidx = ceil ;lidx < freed_blks.size();++lidx){
		hd					= freed_blks[lidx];
		if(hd == EMPTY_LIST) continue;//no list with this size

		asrt(hd+3<mem_sz);
		//remove at front
		uint32_t real_sz	= delFrBlkSize(hd); 	asrt(delFrBlkBegFld(hd)==FREED_BLOCK);
		uint32_t end		= hd + real_sz; 	asrt(delBkBlkEndFld(end)==FREED_BLOCK);
		uint32_t next		= delFrBlkNext(hd);
		freed_blks[lidx]	= next;				asrt(frgmnts_sum>=real_sz);
		frgmnts_sum			-= real_sz;

		if(next != EMPTY_LIST){asrt(delFrBlkBegFld(next)==FREED_BLOCK); asrt(delFrBlkPrev(next)==hd);
			delFrBlkPrev(next)= EMPTY_LIST;}

		asrt(real_sz>=sz);
		//maybe split free memory
		if(real_sz > sz){
			freeUnextendable(hd+sz,end);}

		delClearControlFlags(hd,hd+sz); LOG3("2. Alloc mem from " << hd << " to " << hd+sz<< " mem_sz is " << mem_sz << " mem_end is " << mem_end <<" id " << ID);
		REPRO('A',sz << " at " << hd); //bool ok = ensureUnused(hd,sz); asrt(ok);
		return Ptr32(hd);
	}

	uint64_t req 	= (uint64_t)mem_sz + (uint64_t)sz;
	if(req >(uint64_t)mem_end){
		LOG2("Extend mem with size " << mem_sz << " with fragmentation " << ((double)frgmnts_sum*100.0)/(double)mem_sz<< " id " << ID);
		increaseTo(req+(req>>1),req);
	}

	hd		= mem_sz;
	mem_sz 	+= sz; asrt(mem_sz <= mem_end);


	delClearControlFlags(hd,hd+sz); LOG3("3. Alloc mem from " << hd << " to " << hd+sz<< " mem_sz is " << mem_sz << " mem_end is " << mem_end <<" id " << ID); //asrt(mem[hd] != UN_INIT);
	REPRO('A',sz << " at " << hd); //bool ok = ensureUnused(hd,sz); asrt(ok);
	return Ptr32(hd);
}

void MemPoolA::hardClear(){
	////LOG_ALWAYS("hardClear memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);
	if(mem != NULL){asrt(mem_end != 0);
		free(mem);
		mem			= NULL;
		mem_end		= 0;
		mem_sz		= 0;
		freed_blks	.clear();
		freed_trns	.clear();
	}
	REPRO('H'," "); //dbg_used.clear();
}

void MemPoolA::increaseTo(uint64_t req, uint64_t min){
	////LOG_ALWAYS("increaseTo memsize is " << mem_sz << " mem_end is " << mem_end << " mem pointer is " << (uint64_t)mem << " id " << ID);

	asrt(req > mem_end); asrt(mem_sz<=mem_end);
	if(req>=FREED_BLOCK){
		req = FREED_BLOCK-2;
		if(min>req) goto BAD_ALLOC;}
	asrt(min <= req);

	while(true){
		uint32_t* nmem = (uint32_t*)malloc(req*sizeof(uint32_t));

		if(nmem != NULL){

			if(mem!=NULL){
				memcpy(nmem,mem,mem_sz* sizeof(uint32_t));
				free(mem);
			}else{asrt(mem_sz==0);}

			//memset(nmem+mem_sz,0,(req-mem_sz)*sizeof(uint32_t)); //dbg_used.resize(mem_end,max32);

			mem_end = req;
			mem		= nmem;
			return;
		}

		if(req<=min) goto BAD_ALLOC;

		req = (min>>1) + (req>>1);
	}

	BAD_ALLOC:;
	////LOG_ALWAYS("Out of mem in mempool at size " << req<< " id " << ID);
	exit(EXIT_FAILURE);
}



#endif /* MEMPOOL_HPP_ */
